"""
This file contains the menu items that should be loaded into the main menu.
All of them must inherit from MenuItem
"""
# pylint: disable=missing-class-docstring,too-few-public-methods,import-outside-toplevel
import asyncio
import traceback
import sys

from colorama import Style, Fore

from .menu_item import MenuItem
from .constants import POSITIVE_RESPONSES
from .log_downloader import CkeyLogDownloader, RoundLogDownloader, RoundListLogDownloader

try:
    from .slur_detector import SlurDetector
except FileNotFoundError:
    # Bit of a hack but it does the job
    print(traceback.format_exc().replace("FileNotFoundError:", f"{Fore.RED}FileNotFoundError:") + Fore.RESET)
    print(f"{Style.DIM}Press return to exit...{Style.RESET_ALL}", end='')
    input()
    sys.exit(1)


class CkeySingleItem(MenuItem):
    weight = 2
    name = "ckey log downloader"
    description = "Download someone's say history (and more!)"

    def run(self):
        downloader = CkeyLogDownloader.interactive()
        asyncio.run(downloader.process_and_write())


class SlurDetectorSingleItem(MenuItem):
    name = "slur detector"
    description = "Run slur detection on a file"

    def run(self):
        from .slur_detector.__main__ import main
        main()


class CkeyAndSlurItem(MenuItem):
    weight = 3
    name = "ckey log slur detector"
    description = "Run slur detection on someone's say logs"

    def run(self):
        downloader = CkeyLogDownloader.interactive()
        asyncio.run(downloader.process_and_write())
        print()  # Newline
        slurs = SlurDetector.from_file(downloader.output_path)
        slurs.print_results()


class RoundSingleItem(MenuItem):
    weight = 4
    name = "round log downloader"
    description = "Download logs from a range of rounds"

    def run(self):
        downloader = RoundLogDownloader.interactive()
        asyncio.run(downloader.process_and_write())


class RoundAndSlurItem(MenuItem):
    name = "round slur detector"
    description = "Run slur detection on a range of rounds"

    def run(self):
        downloader = RoundLogDownloader.interactive()
        asyncio.run(downloader.process_and_write())
        print()  # Newline
        slurs = SlurDetector.from_file(downloader.output_path)
        slurs.print_results()


class CentComItem(MenuItem):
    name = "CentCom"
    description = "Search the CentCom ban database for ckeys"

    def run(self):
        from .centcom.__main__ import main
        main()


class UserExistsItem(MenuItem):
    name = "BYOND user exists"
    description = "Check if users on a list exist or not"

    def run(self):
        from .byond.__main__ import main
        main()


class TokenTestServiceItem(MenuItem):
    name = "Token test service"
    description = "The one-stop shop for TG13 token testing"

    def run(self):
        from .auth.__main__ import main
        main()


class PlayedTogetherItem(MenuItem):
    name = "Rounds played together"
    description = "Tells you the rounds two (or more) people have all played in"

    def run(self):  # noqa: C901
        while True:
            try:
                number_of_rounds = int(input("How many rounds? "))
                break
            except ValueError:
                print("That doesn't seem to be a number...")

        from .scrubby import GetReceipts, ScrubbyException
        print("Please enter the desired ckeys. Leave empty to stop")
        receipts_collection = []
        ckeys = []
        while ckey := input():
            try:
                ckeys.append(ckey)
                receipts = asyncio.run(GetReceipts(ckey, number_of_rounds, False))
                receipts_collection.append(receipts)
            except ScrubbyException:
                print("Seems like that ckey couldn't be found! Check your spelling and try again")
        print("Calculating...")
        round_set = set(rd.roundID for rd in receipts_collection[0])
        if len(receipts_collection) == 1:
            print("Seems like there's only one person here, here's the rounds they played in:")
            print(', '.join(str(x) for x in round_set))
            return

        for receipts in receipts_collection[1:]:
            round_set = round_set & set(rd.roundID for rd in receipts)

        print("Here are your stats:")
        print("I looked for the ckeys", ', '.join(ckeys))
        print(f"Out of {number_of_rounds} rounds, they played " +
              f"{Fore.GREEN}{len(round_set) / number_of_rounds * 100}%{Fore.RESET} together")
        print(f"Those rounds were:{Fore.GREEN}", ', '.join(str(x) for x in round_set) or "none!", Fore.RESET)

        if not round_set:
            return
        print(f"Would you like to download these rounds? [y/{Style.BRIGHT}N{Style.NORMAL}] ", end="")
        if not input().strip() in POSITIVE_RESPONSES:
            return
        downloader = RoundListLogDownloader(round_set)
        downloader.try_authenticate_interactive()
        asyncio.run(downloader.process_and_write())
        print("Saved as", downloader.output_path)


class LogBuddyItem(MenuItem):
    weight = 1
    name = "LogBuddy"
    description = "Run LogBuddy"

    def run(self):
        from .log_buddy.__main__ import main
        main()


class RoundListDownloaderItem(MenuItem):
    name = "Round list downloader"
    description = "Download a comma separated list of rounds"

    def run(self):
        downloader = RoundListLogDownloader.interactive()
        asyncio.run(downloader.process_and_write())


class SuspiciousAccessDownloaderItem(MenuItem):
    name = "Suspicious downloader"
    description = "Downloads a range of suspicious access logs"

    def run(self):
        from .auth.tg import interactive as interactive_auth
        print("Input boundry rounds, separated by a space: ")
        while True:
            rounds = input("> ").split(' ', 1)
            if len(rounds) != 2:
                print("Please input two rounds.")
                continue
            if rounds[0].isnumeric() and rounds[1].isnumeric():
                break
            print("Those don't seem to be numbers, please try again")
        downloader = RoundLogDownloader(int(rounds[0]), int(rounds[1]))
        downloader.files = ["suspicious_logins.log"]
        downloader.silent = True
        interactive_auth()
        asyncio.run(downloader.process_and_write())


class JobCounterItem(MenuItem):
    name = "Job counter"
    description = "Display what jobs someone played"

    def run(self):
        from .scrubby import GetReceipts, ScrubbyException
        ckey = input("Ckey: ")
        while True:
            try:
                number_of_rounds = int(input("How many rounds? "))
                break
            except ValueError:
                print("That doesn't seem to be a number...")
        try:
            receipts = asyncio.run(GetReceipts(ckey, number_of_rounds, True))
        except ScrubbyException:
            print("Seems like that ckey couldn't be found! Check your spelling and try again")
            return

        print("Calculating...")
        from collections import Counter
        number_of_rounds = len(receipts)
        print(f"Out of {Fore.GREEN}{number_of_rounds}{Fore.RESET} last round{'s' if number_of_rounds != 1 else ''} " +
              f"{Fore.GREEN}{ckey}{Fore.RESET} played in, they played the following jobs:\n")
        for job, count in Counter(x.job for x in receipts).items():
            print(f"{job:>24}:    {count / number_of_rounds * 100:05.2f}% ({count})")
        antag = sum(x.antagonist for x in receipts)
        suicide = sum(x.roundStartSuicide for x in receipts)
        # roller = sum(x.roundStartSuicide and x.antagonist for x in receipts)
        print(f"\nOut of those, they were an antagonist {antag / number_of_rounds * 100:.2f}% " +
              f"of the time ({antag} round{'s' if antag != 1 else ''}), and round-start suicided " +
              f"{suicide} time{'s' if suicide != 1 else ''}.")
        # if roller:
        #     print(f"{Fore.RED}Out of those {suicide} suicides, they got an antagonist " +
        #           f"{roller} time{'s' if suicide != 1 else ''}!{Fore.RESET}")
        # else:
        #     print(f"{Fore.GREEN}They never round-start suicided and got antagonist.{Fore.RESET}")


# Commented out because it sucked too much.
# class TextSimilarityItem(MenuItem):
#     name = "Text similarity"
#     description = "Determine how similarly two people tend to type"

#     def run(self):
#         from .util import jaccard_similarity
#         from .byond import user_exists, canonicalize
#         print(f"{Fore.YELLOW}This uses a metric called Jaccard similarity. Results may vary.{Fore.RESET}")
#         while True:
#             ckey1 = canonicalize(input("Ckey 1: "))
#             if user_exists(ckey1):
#                 break
#             print(f"{ckey1} does not seem to exist, try again?")
#         while True:
#             ckey2 = canonicalize(input("Ckey 2: "))
#             if user_exists(ckey2):
#                 break
#             print(f"{ckey2} does not seem to exist, try again?")
#         from .log_buddy import LogFile, LogType
#         downloader = CkeyLogDownloader(ckey1, only_played=True, number_of_rounds=20)
#         downloader.files = ["game.txt"]
#         downloader.filter_logs = True
#         downloader.output_only_log_line = True
#         asyncio.run(downloader.process_and_write())
#         log_file = LogFile.from_file(downloader.output_path)
#         log_file.filter_ckeys(ckey1)
#         log_file.filter_by_type([LogType.SAY])
#         person1_words = []
#         for log in log_file.logs:
#             person1_words.extend(log.text.split())
#         downloader = CkeyLogDownloader(ckey2, only_played=True, number_of_rounds=20)
#         downloader.files = ["game.txt"]
#         downloader.filter_logs = True
#         downloader.output_only_log_line = True
#         asyncio.run(downloader.process_and_write())
#         log_file = LogFile.from_file(downloader.output_path)
#         log_file.filter_ckeys(ckey2)
#         log_file.filter_by_type([LogType.SAY])
#         person2_words = []
#         for log in log_file.logs:
#             person2_words.extend(log.text.split())
#         del log_file

#         print(f"\n{Fore.GREEN}Jaccard similarity index:",
#               jaccard_similarity(person1_words, person2_words),
#               Fore.RESET)



def jaccard_similarity(text1: list[str], text2: list[str]):
    """Returns the jaccard similarity between two lists """
    intersection_cardinality = len(set.intersection(*[set(text1), set(text2)]))
    union_cardinality = len(set.union(*[set(text1), set(text2)]))
    return intersection_cardinality/float(union_cardinality)



from ..constants import __version__

ALL_LOGS_WE_PARSE = [
    "game.txt",
    "attack.txt",
    "pda.txt",
    "silicon.txt",
    "mecha.txt",
    "virus.txt",
    "telecomms.txt",
    "uplink.txt",
    "shuttle.txt"
]
HELP_LINK = "https://github.com/RigglePrime/SS13-tools/blob/master/ss13_tools/log_buddy/README.md"
SHAMELESS = f"## Created using SS13-Tools LogBuddy {__version__} https://github.com/RigglePrime/SS13-Tools\n"
ERRORED_FILE = "errored.log"
MAX_SUPPORTED_LOG_VERSION = "1.0.0"

LOG_COLOUR_SCARLET = 124
LOG_COLOUR_RED = 167
LOG_COLOUR_EMERALD = 35
LOG_COLOUR_PERIWINKLE = 69
LOG_COLOUR_PINK = 169
LOG_COLOUR_GRAY = 245
LOG_COLOUR_PASTEL_CYAN = 159
LOG_COLOUR_SUNSET = 209
LOG_COLOUR_PASTEL_ORANGE = 223
LOG_COLOUR_AMETHYST = 141
LOG_COLOUR_OCEAN = 27



#!/usr/bin/env python3
from __future__ import annotations

import asyncio
import os
from enum import Enum
import traceback
from typing import Annotated, Iterable, Union, Literal
from html import unescape as html_unescape
from itertools import repeat, chain

from tqdm import tqdm

from .log import Log, LogType
from .constants import ALL_LOGS_WE_PARSE, ERRORED_FILE, SHAMELESS
from ..__version__ import __version__
from ..log_downloader import RoundLogDownloader, RoundListLogDownloader, CkeyLogDownloader
from ..scrubby import get_round_source_url


HEARING_RANGE = 9


class NotSortableException(Exception):
    """Hey, I can't sort this!"""


class InvalidType(Exception):
    """What kind of type is this..."""


class UnsupportedLogTypeException(Exception):
    """Strange log type you have there"""


class LogParserException(Exception):
    """Strange log type you have there"""


class LogFileType(Enum):
    """Enum of possible log file types"""
    UNKNOWN = 0
    COLLATED = 1
    GAME = 2
    ATTACK = 3
    PDA = 4
    SILICON = 5
    MECHA = 6
    VIRUS = 7
    TELECOMMS = 8
    UPLINK = 9
    SHUTTLE = 10
    TGUI = 11

    @staticmethod
    def parse_log_file_type(string: str):
        """Tries to parse a log type from a string"""
        try:
            return LogFileType[string.upper()]
        except KeyError:
            return LogFileType.UNKNOWN

    @staticmethod
    def list():
        """Returns all possible log file types"""
        return list(LogFileType)


class LogFile:
    """An object representing a log file. Most functions use `self.work_set`, original logs sorted in `self.logs`.

    Parameters:
    `logs` (list[str]): list of log lines
    `type` (LogFileType): type of the log file
    `verbose` (bool): toggles verbose mode
    `quiet` (bool): toggles quiet mode

    Examples:

    `log_file = LogFile()` # Empty log file, useful for combining more later using `collate`,

    `log_file = LogFile(open("game.log"), LogFileType.UNKNOWN)`,

    `log_file = LogFile(["logline 1", "log line 2", "log line 3"])\
    # NOTE: must be a valid log or the parser will raise an exception`
    """
    round_id: Annotated[int, "Stores the round ID. If unknown, it will equal -1"]
    unfiltered_logs: Annotated[list[Log], "Stores a list of all logs"]
    logs: Annotated[list[Log], "Stores a list of filtered logs"]
    who: Annotated[list[str], "Stores a list of all connected ckeys"]
    sortable: bool
    log_source: Annotated[str, "Source of the logs (if available)"]

    def __init__(self, logs: Iterable[str] = None, log_type: LogFileType = LogFileType.UNKNOWN,
                 verbose: bool = False, quiet: bool = False) -> None:
        if verbose and quiet:
            print("Really? You want me to be silent and verbose? Those are mutually exclusive you know")
        self.round_id = -1
        self.unfiltered_logs = []
        self.logs = []
        self.who = []
        self.sortable = True
        self.log_type = log_type
        self.log_source = None

        if not logs:
            return

        self.__parse_logs(logs, verbose=verbose, quiet=quiet)
        self.unfiltered_logs.sort(key=lambda log: log.time)
        self.logs = self.unfiltered_logs

    def __parse_logs(self, logs: Iterable[str], verbose: bool = False, quiet: bool = False):
        errored = []
        pbar = tqdm(logs)
        for line in pbar:
            line = line.strip()
            if not line or line == "- -------------------------" or line == '-' \
                    or "] Starting up round ID " in line or line.startswith("##"):
                continue
            try:
                self.__parse_one_line(line)
            except Exception as exception:  # pylint: disable=broad-exception-caught
                errored.append(line)
                pbar.clear()
                if not quiet:
                    print(f"Could not be parsed: '{line}', with the reason:", exception)
                if verbose:
                    traceback.print_exc()
                pbar.display()
        if errored:
            with open(ERRORED_FILE, 'a+', encoding="utf-8") as file:
                file.write("## If you see this, please share it with Riggle.\n")
                file.write("## ")
                file.write(__version__)
                file.write("\n\n")
                file.writelines(chain.from_iterable(zip(errored, repeat("\n"))))

    def __parse_one_line(self, line: str):
        if line.startswith("-censored"):
            return  # Skip censored lines

        # VOTE is split into multiple lines, so account for that
        if line.startswith("- <b>") and self.unfiltered_logs and self.unfiltered_logs[-1].log_type == LogType.VOTE:
            self.unfiltered_logs[-1].text += ", " + html_unescape(line.replace("- <b>", "").replace("</b>", ""))
            return

        # Priority announcements (and others like it) sometimes do this
        if line.startswith("- ") and self.unfiltered_logs:
            # Don't actually insert a new line
            line = self.unfiltered_logs[-1].raw_line + "\\n" + line.replace("- ", "")
            # Remove the incomplete entry (so we can parse location too!)
            self.unfiltered_logs.pop()
        log = Log(line)
        self.unfiltered_logs.append(log)
        if log.agent and log.agent.ckey and log.agent.ckey not in self.who:
            self.who.append(log.agent.ckey)

    def add_log(self, log: Log, reset_workset: bool = True, sort: bool = True) -> None:
        """Appends a log entry to the end.

        Parameters:
        `log` (Log): the Log object to be added
        `reset_workset` (bool): if we should also reset the working set
        `sort` (bool): if we should also sort the logs afterwards

        NOTE: the log variable MUST be of type Log

        Returns None
        """
        if not isinstance(log, Log):
            raise InvalidType(f"Type Log required but type {str(type(log))} was found")
        self.unfiltered_logs.append(log)
        if reset_workset:
            self.reset_work_set()
        if sort:
            self.sort()

    def add_logs(self, logs: list[Log], reset_workset: bool = True, sort: bool = True) -> None:
        """Appends a list of log entries to the end.

        Parameters:
        logs (list[Log]): the Log objects list to be added
        `reset_workset` (bool): if we should also reset the working set
        `sort` (bool): if we should also sort the logs afterwards

        Returns None
        """
        self.unfiltered_logs.extend(logs)
        if reset_workset:
            self.reset_work_set()
        if sort:
            self.sort()

    def sort(self) -> None:
        """Sorts the current work set, using the time at which the log was added, descending

        Example call: `my_logs.sort()`

        Returns None"""
        if not self.sortable:
            raise NotSortableException("Not enough information to sort the logs")
        self.logs.sort(key=lambda log: log.time)

    def collate(self, logfile: LogFile) -> None:
        """Collates (extends, adds together) two LogFile objects and changes the LogFileType to COLLATED.
        The result is stored in the the object this was called on. A call to this function will reset the
        current work set.

        Parameters:
        `logfile` (LogFile): the LogFile object you want to combine

        Example:
        `my_logs = LogFile()`
        `my_logs.collate(LogFile.from_file("game.txt"))`
        `my_logs.collate(LogFile.from_file("attack.txt"))`

        Returns `None`
        """
        self.add_logs(logfile.unfiltered_logs, sort=True)
        self.log_type = LogFileType.COLLATED
        self.who.extend(logfile.who)
        self.who = list(set(self.who))  # Remove duplicates
        self.who.sort()
        self.logs = self.unfiltered_logs

    def filter_ckeys(self, *ckeys: str, source_only: bool = False) -> None:
        """Removes all logs in which the specified ckeys are not present, saving the result
        in self.work_set. Works much like Notepad++, but only counts the agent (actor, the
        one who performed the action). See `filter_strings` for a function like Notepad++ bookmark

        Parameters:
        `ckeys` (tuple[str, ...]): ckeys to filter

        Example call: `my_logs.filter_ckeys("ckey1", "ckey2")` (as many or little ckeys as you want)

        Returns `None`"""
        filtered = []
        ckeys = [x.casefold() for x in ckeys]
        for log in self.logs:
            if (log.agent and log.agent.ckey and log.agent.ckey.casefold() in ckeys) or \
               (not source_only and log.patient and log.patient.ckey and log.patient.ckey.casefold() in ckeys):
                filtered.append(log)
        if not filtered:
            print("Operation completed with empty set. Aborting.")
            return
        self.logs = filtered

    def filter_strings(self, *strings: str, case_sensitive: bool = False, additive: bool = False) -> None:
        """Removes all logs in which the specified strings are not present, saving them in
        `self.work_set`. Works exactly like Notepad++ bookmark

        Parameters:
        `strings` (tuple[str, ...]): strings to filter
        `case_sensitive` (bool): toggles case sensitivity

        Example calls: `my_logs.filter_strings("Hi!")`
        `my_logs.filter_strings("attacked", "injected", "I hate you")`
        `my_logs.filter_strings("racial slur", case_sensitive=True)` (as many strings as you want)

        Returns `None`"""
        filtered = []
        logs_it = self.unfiltered_logs if additive else self.logs
        for log in logs_it:
            raw_line = log.raw_line
            if not case_sensitive:
                raw_line = raw_line.casefold()
            for string in strings:
                if not case_sensitive:
                    string = string.casefold()
                if string in raw_line:
                    filtered.append(log)
                break
        if not filtered:
            print("Operation completed with empty set. Aborting.")
            return
        if additive:
            self.logs.extend(filtered)
            self.logs = list(set(self.logs))  # Make sure they're unique
            self.sort()
        else:
            self.logs = filtered

    def filter_strings_case_sensitive(self, *strings: str) -> None:
        """Shorter for `filter_strings(*strings, case_sensitive = True)`"""
        self.filter_strings(*strings, case_sensitive=True)

    def filter_heard(self, *ckeys: str, walking_error: int = 4) -> None:
        """Removes all log entries which could not have been heard by the specified ckey(s)
        (very much in alpha) and stores the remaining lines in `self.work_set`

        Parameters:
        `ckeys` (tuple[str, ...]): desired ckey(s)
        `walking_error` (int): added to hearing range to account for the lack of logs

        Example call: `my_logs.filter_heard("ckey")`

        Returns `None`"""
        final = set()
        for ckey in ckeys:
            final.update(self._get_only_heard(ckey, walking_error=walking_error))
        self.logs = list(final)
        self.sort()

    def filter_conversation(self, *ckeys: str, walking_error: int = 4) -> None:
        """Tries to get a conversation between multiple parties, excluding what they would and would not hear as a group.
        Saves the result in `self.work_set`

        Parameters:
        `ckeys` (tuple[str, ...]): ckeys to use for sorting
        `walking_error` (int): added to hearing range to account for the lack of logs

        Example call: `my_logs.filter_conversation("ckey1", "ckey2", "ckey3")` (as many or little ckeys as you want)

        Returns None"""
        self.filter_ckeys(*ckeys, source_only=False)
        final = self._get_only_heard(ckeys[0], walking_error=walking_error)
        for ckey in ckeys[1:]:
            final.intersection_update(self._get_only_heard(ckey, walking_error=walking_error))

        if not final:
            print("Operation completed with empty set. Aborting.")
            return
        self.logs = list(final)
        self.sort()

    def reset_work_set(self):
        """Removes all filters; sets the working set to be equal to all logs

        Example call: my_logs.reset_work_set()"""
        self.logs = self.unfiltered_logs

    def _get_only_heard(self, ckey: str, logs_we_care_about: Union[list[LogType],
                        Literal["ALL"]] = "ALL", walking_error: int = 4) -> set[Log]:
        """Removes all log entries which could not have been heard by the specified ckey (very much in alpha).
        Uses logs from `self.work_set`

        Parameters:
        `ckey` (str): ckeys to use
        `logs_we_care_about` (list[LogType])
        `walking_error` (int): added to hearing range to account for the lack of logs

        Example calls: `my_logs.get_only_heard("ckey")`
        `my_logs.get_only_heard("ckey", "ALL")`
        `my_logs.get_only_heard("ckey", [LogType.SAY, LogType.WHISPER])`

        Returns `set[Log]`"""
        self.sort()
        # Adjust for error created by lack of logs
        hearing_range = HEARING_RANGE + walking_error
        filtered = set()
        cur_loc = (0, 0, 0)
        last_loc = cur_loc
        # Iterate through unfiltered logs to actually get the data we want
        for log in self.unfiltered_logs:
            # Check for ckey. If our target was included in the action we can safely assume they saw it
            if (log.agent and ckey == log.agent.ckey) or\
               (log.patient and ckey == log.patient.ckey) or\
               (log.text and f"{ckey}/(" in log.text):
                # If there's a location attached, update it
                if log.location:
                    last_loc = cur_loc
                    cur_loc = log.location
                filtered.add(log)
                continue
            # If our target didn't participate, we need to check how far away it happened

            # Check z-level, if they differ save location and continue
            if cur_loc[2] != last_loc[2]:
                continue
            # Filter logs that we don't care about but still use their location
            if logs_we_care_about and (logs_we_care_about != "ALL"):
                continue
            if isinstance(logs_we_care_about, list) and log.log_type not in logs_we_care_about:
                continue
            # Skip logs with no location data available
            if not log.location:
                continue
            # Calculate distance
            # if sqrt(pow(cur_loc[0] - log.location[0], 2) + pow(cur_loc[1] - log.location[1], 2)) - hearing_range < 0:
            if abs(cur_loc[0] - log.location[0]) - hearing_range < 0 and abs(cur_loc[1] - log.location[1]) - hearing_range < 0:
                filtered.add(log)
            # You (almost) always hear tcomms
            elif log.log_type == LogType.TCOMMS:
                filtered.add(log)

        # Intersect with currently filtered logs so we don't reset filters
        return filtered & set(self.logs)

    def filter_by_location_name(self, location_name: str, exact: bool = False) -> None:
        """Removes all logs that did not happen in the specified location,
        and stores the result in the work set.

        Parameters:
        `location_name` (str): the name of the location, case insensitive
        `exact` (bool): should the strings exactly match, or just contain?

        Example call: my_logs.filter_by_location_name("Bar")

        Returns `None`"""
        filtered = []
        location_name = location_name.casefold()
        for log in self.logs:
            if not log.location_name:
                continue
            if (not exact and location_name in log.location_name.casefold()) or \
               (location_name == log.location_name.casefold()):
                filtered.append(log)
        if not filtered:
            print("Operation completed with empty set. Aborting.")
            return
        self.logs = filtered

    def filter_by_radius(self, location: tuple[int, int, int], radius: int, exclude_locationless: bool = True) -> None:
        """Removes all logs that did not happen in the specified radius around the location,
        and stores the result in the work set.

        Parameters:
        `location` (tuple[int, int, int]): the location
        `radius` (int): the radius

        Example call: `my_logs.filter_by_radius((32, 41, 2), 5)`

        Returns None"""
        filtered = []
        for log in self.logs:
            if not log.location:
                if not exclude_locationless:
                    filtered.append(log)
                continue
            # Z level must match
            if log.location[2] != location[2]:
                continue
            if abs(location[0] - log.location[0]) - radius < 0 and abs(location[1] - log.location[1]) - radius < 0:
                filtered.append(log)
        if not filtered:
            print("Operation completed with empty set. Aborting.")
            return
        self.logs = filtered

    def filter_by_type(self, include: Iterable[LogType] = None, exclude: Iterable[LogType] = None):
        """Only keeps (or removes) logs lines of the specified type.

        Parameters:
        `include` (tuple[LogType]): log types to include
        `exclude` (tuple[LogType]): log types to exclude

        If the same type is seen in include and exclude, it will be excluded.
        To get a list of all available LogTypes call `LogType.list()`

        Example calls:
        `my_logs.filter_by_type((LogType.OOC,))` (the first argument counts as include)
        `my_logs.filter_by_type(include=(LogType.SAY,))`
        `my_logs.filter_by_type(exclude=(LogType.TCOMMS,))`
        """

        if include:
            filter_for = set(LogType) & set(include)
        else:
            filter_for = set(LogType) - set(exclude)
        if not filter_for:
            print("Nothing to filter for!")
            return
        filtered = []
        for log in self.logs:
            if log.log_type in filter_for:
                filtered.append(log)
        if not filtered:
            print("Operation completed with empty set. Aborting.")
            return
        self.logs = filtered

    def print_working(self) -> None:
        """Prints working set to the console

        Example call: `my_logs.print_working()`

        Returns `None`"""
        if not self.logs:
            print("Working set empty")
            return
        for log in self.logs:
            print(log.pretty())

    def head(self, number: int = 10) -> None:
        """Prints the first few lines of the working set to the console.

        Parameters:
        `n` (int): number to print, defaults to 10

        Example call: `my_logs.head()`

        Returns `None`"""
        if not self.logs:
            print("Working set empty")
            return
        for log in self.logs[:number]:
            print(log.pretty())

    def tail(self, number: int = 10) -> None:
        """Prints the last few lines of the working set to the console.

        Parameters:
        `n` (int): number to print, defaults to 10

        Example call: `my_logs.tail()`

        Returns `None`"""
        if not self.logs:
            print("Working set empty")
            return
        for log in self.logs[-number:]:
            print(log.pretty())

    def write_working_to_file(self, filename: str, force_overwrite: bool = False) -> None:
        """Writes current `self.work_set` to the desired file.

        Parameters:
        `filename` (str): name of the file to write to (overwrites everything)

        Example call: `my_logs.write_working_to_file("logs.txt")`

        Returns None"""
        if not force_overwrite and os.path.exists(filename):
            print("Seems like there's already a file at that location! Overwrite? [y/N] ", end='')
            if input().strip().lower() != 'y':
                print("Nothing was saved.")
                return
            print("Okay then, overwriting")
        with open(filename, "w", encoding="utf-8") as file:
            for log in self.logs:
                file.write(str(log) + "\n")
            file.write(SHAMELESS)
            if self.log_source:
                file.write(f"## Logs acquired from {self.log_source}")

    def __len__(self) -> int:
        """Returns the length of the logs array"""
        return self.logs.__len__()

    def __getitem__(self, key):
        """Access logs like an array, delegates to the underlying logs list"""
        return self.logs.__getitem__(key)

    @staticmethod
    def from_file(filename: str, log_type: LogFileType = None, verbose: bool = False, quiet: bool = False) -> LogFile:
        """Parses the specified log file

        Parameters:
        `filename` (str): name (and location) of the desired file
        `type` (LogFileType): type of the log. Use if you want to override log type detection
        (optional, defaults to LogFileType.UNKNOWN)
        `verbose` (bool): toggle verbose mode (False by default)
        `quiet` (bool): toggle quiet mode (False by default)

        Example call: `my_logs = LogFile.from_file("game.txt")`

        Returns LogFile"""
        if filename.endswith(".html"):
            raise UnsupportedLogTypeException(f"{filename} does not seem to be supported")
        if not log_type and "." in filename:
            log_type = LogFileType.parse_log_file_type(filename.split(".", 1)[0])
        with open(filename, "r", encoding="utf-8") as file:
            return LogFile(file, log_type, verbose, quiet)

    @staticmethod
    def from_folder(folder: str, verbose: bool = False, quiet: bool = False) -> LogFile:
        """Parses all log files in a folder, combining them into a single file

        Parameters:
        `filename` (str): name (and location) of the desired folder
        `verbose` (bool): toggle verbose mode (False by default)
        `quiet` (bool): toggle quiet mode (False by default)

        Example call: `my_logs = LogFile.from_file("game.txt")`

        Returns `LogFile`"""
        if not os.path.isdir(folder):
            raise FileNotFoundError("Is not a folder")
        folder = folder.replace("\\", "/")
        if folder[-1] != "/":
            folder += "/"
        log_collection = LogFile()
        for file in os.listdir(folder):
            if not quiet:
                print("Parsing", file)
            try:
                log_collection.collate(LogFile.from_file(folder + file, verbose=verbose, quiet=quiet))
            except UnsupportedLogTypeException:
                if not quiet:
                    print(f"{file} isn't supported, skipping")
                continue
        return log_collection

    @staticmethod
    def from_round_id(round_id: int, logs_we_care_about: list[str] = None) -> LogFile:
        """Downloads multiple files from a round ID.

        Parameters:
        `round_id` (int): round to download
        `logs_we_care_about` (list[str]): list of strings, containing the file names.
        For example: `["game.txt", "attack.txt"]`. This defaults to all supported files.

        Example call: `my_logs = LogFile.from_round_id(185556)`

        Returns `LogFile`"""
        # Should be all supported log types as a default. Don't forget to update this list! (you will)
        if not logs_we_care_about:
            logs_we_care_about = ALL_LOGS_WE_PARSE.copy()

        downloader = RoundLogDownloader(round_id, round_id, f"{round_id}.log")
        downloader.output_only_log_line = True
        downloader.files = logs_we_care_about
        downloader.try_authenticate_interactive()
        asyncio.run(downloader.process_and_write())
        log_collection = LogFile.from_file(downloader.output_path)
        log_collection.log_type = LogFileType.COLLATED
        # Sort the logs
        log_collection.log_source = get_round_source_url(round_id=round_id)
        log_collection.write_working_to_file(downloader.output_path, force_overwrite=True)
        return log_collection

    @staticmethod
    def from_round_range(start_round_id: int, end_round_id: int, logs_we_care_about: list[str] = None) -> LogFile:
        """Downloads multiple rounds worth of data.

        Parameters:
        `start_round_id` (int): first round to download
        `end_round_id` (int): last round to download (inclusive)
        `logs_we_care_about` (list[str]): list of strings, containing the file names.
        For example: `["game.txt", "attack.txt"]`. This defaults to all supported files.

        Example call: `my_logs = LogFile.from_multiple_rounds(185556, 191100)`

        Returns `LogFile`"""
        # Should be all supported log types as a default. Don't forget to update this list! (you will)
        if not logs_we_care_about:
            logs_we_care_about = ALL_LOGS_WE_PARSE.copy()

        downloader = RoundLogDownloader(start_round_id, end_round_id)
        downloader.output_only_log_line = True
        downloader.files = logs_we_care_about
        downloader.try_authenticate_interactive()
        asyncio.run(downloader.process_and_write())
        log_collection = LogFile.from_file(downloader.output_path)
        log_collection.log_type = LogFileType.COLLATED
        log_collection.log_source = f"{start_round_id}-{end_round_id}"
        log_collection.write_working_to_file(downloader.output_path, force_overwrite=True)
        return log_collection

    @staticmethod
    def from_round_collection(*rounds: int, logs_we_care_about: list[str] = None) -> LogFile:
        """Downloads multiple rounds worth of data.

        Parameters:
        `*rounds` (int): list of rounds to download
        `logs_we_care_about` (list[str]): list of strings, containing the file names.
        For example: `["game.txt", "attack.txt"]`. This defaults to all supported files.

        Example call: `my_logs = LogFile.from_round_collection(185556, 185558, 185560, ...)`

        Returns `LogFile`"""
        # Should be all supported log types as a default. Don't forget to update this list! (you will)
        if not logs_we_care_about:
            logs_we_care_about = ALL_LOGS_WE_PARSE.copy()

        downloader = RoundListLogDownloader(rounds)
        downloader.output_only_log_line = True
        downloader.files = logs_we_care_about
        downloader.try_authenticate_interactive()
        asyncio.run(downloader.process_and_write())
        log_collection = LogFile.from_file(downloader.output_path)
        log_collection.log_type = LogFileType.COLLATED
        log_collection.log_source = "rounds " + ', '.join(str(x) for x in rounds)
        log_collection.write_working_to_file(downloader.output_path, force_overwrite=True)
        return log_collection

    @staticmethod
    def from_ckey(ckey: str, rounds: int = 20, only_played: bool = False, filter_logs: bool = False,
                  logs_we_care_about: list[str] = None) -> LogFile:
        """Downloads multiple rounds worth of data, where the specified ckey was present.

        Parameters:
        `ckey` (str): the ckey
        `rounds` (int): number of rounds
        `filter` (bool): pre-emptively delete all logs not containing their ckey
        `logs_we_care_about` (list[str]): list of strings, containing the file names.
        For example: `["game.txt", "attack.txt"]`. This defaults to all supported files.

        Example call: `my_logs = LogFile.from_ckey("Riggle")`

        Returns `LogFile`"""
        # Should be all supported log types as a default. Don't forget to update this list! (you will)
        if not logs_we_care_about:
            logs_we_care_about = ALL_LOGS_WE_PARSE.copy()

        downloader = CkeyLogDownloader(ckey, only_played, rounds)
        downloader.output_only_log_line = True
        downloader.files = logs_we_care_about
        downloader.filter_logs = filter_logs
        downloader.try_authenticate_interactive()
        asyncio.run(downloader.process_and_write())
        log_collection = LogFile.from_file(downloader.output_path)
        log_collection.log_type = LogFileType.COLLATED
        log_collection.log_source = f"{rounds} latest rounds that {ckey} played in"
        log_collection.write_working_to_file(downloader.output_path, force_overwrite=True)
        return log_collection


if __name__ == "__main__":
    import sys
    LogFile.from_file(sys.argv[1]).print_working()



from colorama import Fore

from ss13_tools.centcom.ban import get_one


def main():
    """Main"""
    try:
        print("Paste ckeys to search for, one per line (press CTRL + C to stop)\n")
        while True:
            key = input('> ')
            if not key:
                print("Enter a ckey, or CTRL+C to stop")
                continue
            ban_data = get_one(key)

            print(f"{key}:\n")
            if len(ban_data) == 0:
                print("No data")
            for ban in ban_data:
                print(f"{Fore.CYAN}{ban.sourceName}, {ban.sourceRoleplayLevel} RP{Fore.RESET}")
                print(f"Banned on {ban.bannedOn} by {Fore.BLUE}{ban.bannedBy}{Fore.RESET}")
                print("With reason:", ban.reason)
                print("Expires:", ban.expires or f"{Fore.RED}never{Fore.RESET}", "and is",
                      f"{Fore.RED}active{Fore.RESET}" if ban.active else f"{Fore.GREEN}not active{Fore.RESET}")
            print("=========================\n")
    except KeyboardInterrupt:
        print("Bye!")


if __name__ == "__main__":
    main()



# Functions and module names taken directly from Scrubby
from .round_data import PlayerRoundData, RoundInfo
from .CKeyController import GetReceipts, ScrubbyException
from .RoundController import get_round_source_url, get_round_info_from_ids, \
                             get_multiple_round_source_urls, get_multiple_round_json, \
                             get_round_json

__all__ = [
    'PlayerRoundData',
    'RoundInfo',
    'GetReceipts',
    'get_round_source_url',
    'get_round_info_from_ids',
    'get_multiple_round_source_urls',
    'get_multiple_round_json',
    'get_round_json',
    'ScrubbyException',
]



"""Defines our special log magics so people can use short commands"""
# pylint: disable=unused-argument
import re
import os
from functools import wraps

from IPython.core.page import page
from IPython.terminal.magics import Magics, magics_class, line_magic
from IPython.core.error import UsageError, StdinNotImplementedError
from pyperclip import copy

from ..byond import canonicalize
from .log_parser import LogFile, SHAMELESS
from .log import LogType


LOGS_VARIABLE_NAME = 'logs'


# For info on pasting see IPython.terminal.magics.TerminalMagics.paste
@magics_class
class LogMagics(Magics):
    """Stores our custom log magics for ease of use"""

    def _undoable(func):  # pylint: disable=no-self-argument
        """Marks a function as being able to be undone. Apply to magics that modify the logs"""
        @wraps(func)
        def decorator_undoable(self, arg):
            """You should not be seeing this"""
            func(self, arg)  # pylint: disable=not-callable
            self.actions.append((func.__name__, arg))  # pylint: disable=no-member
        return decorator_undoable

    @property
    def logs_var(self) -> LogFile:
        """Stores the logs (main LogFile) variable"""
        return self.shell.user_ns[LOGS_VARIABLE_NAME]

    @logs_var.setter
    def logs_var(self, new_value) -> None:
        self.shell.user_ns[LOGS_VARIABLE_NAME] = new_value

    @line_magic
    def download(self, parameter_s=''):
        """Downloads logs from a round ID (or ckey) and stores it in `logs`

        - Options:
            - c: force the program to interpret your input as a ckey
            - p: only get the round the person played in? (applies only for ckeys)
            - r: the amount of rounds to download (applies only for ckeys)
            - f: pre-emptively delete all logs without our target ckey (applies only for ckeys)
        - `%download 198563`: download round 198563
        - `%dl 198563`: same as above
        - `%download 199563-199999`: download rounds 199500 to 199600 (inclusive). Be careful with this,
        as downloading and loading more than 100 rounds may slow down your computer by a lot.
        - `%download 198563 198565 198569 198570`: download rounds 198563, 198565, 198569 and 198570.
        You can run this command with as many rounds as you want.
        - `%download -p coolckey123` downloads 50 (default) rounds that `coolckey123` played in
        - `%download -r10 coolckey123` downloads 10 rounds that `coolckey123` played in
        - `%download -c 123123123` this option is useful, as `dl` would interpret numbers as a round ID.
        `-c` forces it to interpret it as a ckey
        - `%download -c -r=10 123123123` you can have more than one switch! The `=` is optional
        - `%download -cr=10 123123123` you can group options like this! The `=` is optional. Just remember
        to have the number right next to the character (r). Doing `-rc10` would assign 10 to c (error)
        """
        if not parameter_s:
            raise UsageError(f"No arguments! Usage:\n{self.download.__doc__}")
        opts, args = self.parse_options(parameter_s, 'cpr:f')
        if 'c' not in opts and '-' in args:
            args = args.split('-')
            if len(args) != 2:
                raise UsageError("I don't know what to do with this, please try again")
            try:
                first, last = (int(x) for x in args)
            except ValueError as ex:
                raise UsageError("One of these is not a number, try again") from ex
            self.logs_var = LogFile.from_round_range(first, last)
            return
        if 'c' not in opts and ' ' in args or ',' in args:
            # Split with spaces and commas
            args = re.split(r'[, ]', args)
            try:
                # Prase ints and get rid of empty strings
                round_ids = tuple(int(x) for x in args if x)
            except ValueError as ex:
                raise UsageError("One of those is not a number, please try again") from ex
            self.logs_var = LogFile.from_round_collection(*round_ids)
            return
        if 'c' not in opts and args.isnumeric():
            self.logs_var = LogFile.from_round_id(int(args))
        else:
            rounds = int(opts['r'].lstrip('=')) if 'r' in opts else 50
            filter_logs = 'f' in opts
            only_played = 'p' in opts
            args = canonicalize(args)
            self.logs_var = LogFile.from_ckey(args, rounds=rounds, only_played=only_played, filter_logs=filter_logs)

    @line_magic
    def length(self, parameter_s=''):
        """Prints how many logs we have"""
        print("Current number of logs is", len(self.logs_var))

    @line_magic
    def sort(self, parameter_s=''):
        """Sorts our logs"""
        self.logs_var.sort()
        print("Logs sorted!")

    @_undoable
    @line_magic
    def search_ckey(self, parameter_s=''):
        """Excludes logs that do not contain any of the ckeys

        Example:
           - `%ckey WindowSmasher86`
           - `%search_ckey ckey1 ckey2 ...`
        """
        if not parameter_s:
            raise UsageError(f"Add some ckeys! Usage:\n{self.search_ckey.__doc__}")
        parameter_s = tuple(canonicalize(x) for x in re.split(r'[, ]', parameter_s) if x)
        print("Looking for", ', '.join(parameter_s))
        for ckey in parameter_s:
            if ckey not in self.logs_var.who:
                print(f"{ckey} not found! Ignoring!")
        self.logs_var.filter_ckeys(*parameter_s, source_only=False)

    @_undoable
    @line_magic
    def search_string(self, parameter_s=''):
        """Works just like bookmarking in Notepad++, or CTRL+F multiple times. Case insensitive

        Options:
            - a: append mode. Take from the unfiltered logs and append the results to the current work set
            - c: case sensitive mode
            - r: raw mode. Do not parse any strings, ignore quotation marks

        Example:
            - `%string 'help maint'`
            - `%string -r help maint`
            - `%string security`
            - `%string 'thank you very much!'`
            - `%string -a 'string to add'`: instead of filtering the current filtered logs, append the results (union)
                - Running these would produce the same result: `%string string1 string2` or
                `%string string1`, `%string -a string2`
            - `%string -c 'Case Sensitive'`: enable case sensitive mode
            - `%string -ar Case Sensitive`: same as above (you can group options)
        """
        if not parameter_s:
            raise UsageError("No string to search for!")
        opts, args = self.parse_options(parameter_s.replace('\\"', '\\\\"').replace('"', '"""'), 'acr')
        additive = 'a' in opts
        case_s = 'c' in opts
        if 'r' in opts:
            self.logs_var.filter_strings(args, case_sensitive=case_s, additive=additive)
            return
        args = parse_quoted_string(args)
        print("Searching for strings", ', '.join(args))
        print("Append is", "ON," if additive else "OFF,", "case sensitive mode is",
              "ON" if case_s else "OFF,", "and raw mode is", "ON" if additive else "OFF")
        self.logs_var.filter_strings(*args, case_sensitive=case_s, additive=additive)

    @_undoable
    @line_magic
    def heard(self, parameter_s=''):
        """Gets only what the people could have heard (conversation but it's a union)

        Example:
            - `%heard WindowSmasher86`
            - `%heard ckey1, ckey2`
        """
        if not parameter_s:
            raise UsageError(f"Add some ckeys! Usage:\n{self.heard.__doc__}")
        parameter_s = tuple(canonicalize(x) for x in re.split(r'[, ]', parameter_s) if x)
        print("Looking for", ', '.join(parameter_s))
        for ckey in parameter_s:
            if ckey not in self.logs_var.who:
                print(f"{ckey} not found! Ignoring!")
        self.logs_var.filter_heard(*parameter_s)

    @_undoable
    @line_magic
    def conversation(self, parameter_s=''):
        """Tries to reconstruct a conversation between parties (like heard but an intersection instead)

        Example:
            - `%conversation ckey`
            - `%conversation ckey1 ckey2`
        """
        if not parameter_s:
            print(f"Add some ckeys! Usage:\n{self.conversation.__doc__}")
            return
        parameter_s = tuple(canonicalize(x) for x in re.split(r'[, ]', parameter_s) if x)
        print("Filtering conversation on ckeys", ', '.join(parameter_s))
        for ckey in parameter_s:
            if ckey not in self.logs_var.who:
                print(f"{ckey} not found! Ignoring!")
        self.logs_var.filter_conversation(*parameter_s)

    @line_magic
    def who(self, parameter_s=''):
        """List all ckeys present in the round

        Example:
            - `%who`
        """
        print(self.logs_var.who)

    @line_magic
    def list_locations(self, parameter_s=''):
        """Gets all of the different locations in our current working set

        Example:
            - `%list_locations`
        """
        print({x.location_name for x in self.logs_var.logs})

    @line_magic
    def reset(self, parameter_s=''):
        """Resets the work set"""
        self.logs_var.reset_work_set()
        self.actions.clear()
        print("Filters reset!")

    @_undoable
    @line_magic
    def location(self, parameter_s=''):
        """Filters by location name.

        Options:
            - e: filter for exact name

        Example:
            - %location Medbay
            - %location -e Medbay Central
        """
        if not parameter_s:
            raise UsageError(f"Add a location! Usage:\n{self.location.__doc__}")
        opts, args = self.parse_options(parameter_s, 'e')
        print("Filtering for", parameter_s)
        self.logs_var.filter_by_location_name(args, exact='e' in opts)

    @_undoable
    @line_magic
    def radius(self, parameter_s=''):
        """Tries to reconstruct a conversation between parties

        Example:
            - %radius x y z radius
            - %radius 50 62 2 8
        """
        parameter_s = parameter_s.split(' ')
        if len(parameter_s) != 4:
            print(f"Add some arguments! Usage:\n{self.radius.__doc__}")
            return
        try:
            x, y, z, radius = (int(x) for x in parameter_s)  # pylint: disable=invalid-name
        except ValueError as ex:
            raise UsageError("Could not convert to an integer") from ex
        print(f"Filtering by x={x}, y={y}, z={z}, r={radius}")
        self.logs_var.filter_by_radius((x, y, z), radius)

    @_undoable
    @line_magic
    def type(self, parameter_s=''):
        """Filters by log type"""
        if not parameter_s:
            raise UsageError("No log types!")
        exclude = set(LogType.parse_log_type(x[1:].strip()) for x in re.split(r'[, ]', parameter_s) if x and x[0] == '!')
        include = set(LogType.parse_log_type(x.strip()) for x in re.split(r'[, ]', parameter_s) if x and x[0] != '!')
        if LogType.UNKNOWN in include | exclude:
            print("Unrecognised log type. Available types:")
            print(LogType.list())
            print(f"Example include: %{self.type.__name__} GAME ATTACK")
            print(f"Example exclude: %{self.type.__name__} !SILICON")
            return
        print("Filtering by the following rules:")
        print("Including:", ', '.join(str(x) for x in include))
        print("Excluding:", ', '.join(str(x) for x in exclude))
        self.logs_var.filter_by_type(include=include, exclude=exclude)

    @line_magic
    def print_logs(self, parameter_s=''):
        """Prints our filtered logs"""
        if len(self.logs_var.logs) > 200:
            page("Too many logs, opening pager. Press q to quit, enter to advance one line, space to advance a screen\n" +
                 '\n'.join(log.pretty() for log in self.logs_var.logs) + '\n')
        else:
            self.logs_var.print_working()

    @line_magic
    def head(self, parameter_s=''):
        """Prints our filtered logs, but only the head"""
        try:
            self.logs_var.head(int(parameter_s))
        except ValueError:
            self.logs_var.head()

    @line_magic
    def tail(self, parameter_s=''):
        """Prints our filtered logs, but only the tail"""
        try:
            self.logs_var.tail(int(parameter_s))
        except ValueError:
            self.logs_var.tail()

    @line_magic
    def clear(self, parameter_s=''):
        """Clears the logs, freeing memory"""
        try:
            response = self.shell.ask_yes_no("Are you sure you want to remove all logs? [y/N]", default='n')
        except StdinNotImplementedError:
            print("Logs will not be removed.")
            return
        if not response:
            print("Cancelled")
            return
        self.logs_var = LogFile()
        self.actions.clear()
        print("Logs cleared!")

    @line_magic
    def save_logs(self, parameter_s=''):
        """Saves the working set to a file"""
        if not parameter_s:
            parameter_s = "logs.log"
        print(f"Writing to file {parameter_s}")
        self.logs_var.write_working_to_file(parameter_s)

    @line_magic
    def load_logs(self, parameter_s=''):
        """Opens the file and adds all logs to our current collection"""
        if not parameter_s:
            print("Enter a file name!")
        if not os.path.exists(parameter_s):
            raise UsageError("File does not exist")
        print("Loading from", parameter_s)
        self.logs_var.collate(LogFile.from_file(parameter_s))

    actions = []

    @line_magic
    def undo(self, parameter_s=''):
        """Undo an action! Provide a number to undo multiple times"""
        if not self.actions:
            print("Nothing to undo!")
            return
        if parameter_s and not parameter_s.isnumeric():
            raise UsageError("That's not a number! Run it without one, or provide a number please")
        undo_times = int(parameter_s) if parameter_s else 1
        if undo_times < 1:
            raise UsageError("What the hell? I can't undo less than 1 time...")
        if len(self.actions) < undo_times:
            raise UsageError("Cannot undo that many times!")
        actions = self.actions.copy()[:-undo_times]

        print("Rewinding...")
        self.logs_var.reset_work_set()
        for magic, params in actions:
            print(f"Running %{magic} with params: {params}")
            self.shell.run_line_magic(magic, params)
        self.actions.clear()
        self.actions = actions
        print("Restored")

    @line_magic
    def clip(self, parameter_s=''):
        """Copy the current filtered logs to your clipboard"""
        copy('\n'.join(log.raw_line for log in self.logs_var.logs) + '\n\n' + SHAMELESS)
        print("Copied to clipboard!")


def register_aliases(shell):
    """Adds shorthands for all magics"""
    aliases = [
        ('dl', LogMagics.download.__name__),
        ('l', LogMagics.length.__name__),
        ('ckey', LogMagics.search_ckey.__name__),
        ('string', LogMagics.search_string.__name__),
        ('loc', LogMagics.location.__name__),
        ('s', LogMagics.save_logs.__name__),
        ('p', LogMagics.print_logs.__name__),
    ]
    for alias, magic in aliases:
        shell.magics_manager.register_alias(alias, magic, 'line')


def parse_quoted_string(string: str, sep: str = ' ') -> list[str]:
    """Returns a list of strings, separated by a whitespace"""
    quotes = """"'"""  # Very confusing :)
    sep_len = len(sep)
    cur_quote = None
    cur_marker = 0
    assert sep not in quotes
    length = len(string)
    ret = []
    i = -1  # We need i to start at 0 in the loop
    while (i := i + 1) < length:
        if string[i] == "\\":
            # This next line shortens our string by 1, so no need to increment
            string = string.replace("\\", "", 1)
            length -= 1
            continue
        if not cur_quote and string.startswith(sep, i):
            if cur_marker == i:
                cur_marker = i + 1
                continue
            ret.append(string[cur_marker:i])
            # Set the current marker at the beginning of the next string, and
            # i to the last (since it will get +1'd)
            cur_marker = (i := i + sep_len - 1) + 1  # I LOVE the walrus operator
            continue
        if string[i] not in quotes:
            continue
        if not cur_quote:  # Are we in a quoted string?
            if cur_marker != i:
                ret.append(string[cur_marker:i])
            cur_quote = quotes[quotes.index(string[i])]
            cur_marker = i + 1
            continue
        if cur_quote != string[i]:  # Wrong quote, ignore
            continue
        cur_quote = None
        ret.append(string[cur_marker:i])  # Got one!
        cur_marker = i + 1
    if cur_marker < i:  # Check if there's anything left
        ret.append(string[cur_marker:])
    return ret
